from datetime import datetime
import logging
import webapp2
from webapp2_extras import json
from webob import exc


class BaseError(Exception):
    pass


class InternalServerError(BaseError, exc.HTTPInternalServerError):
    pass


class NotFoundError(BaseError, exc.HTTPNotFound):
    pass


class BadRequestError(BaseError, exc.HTTPBadRequest):
    pass


class ForbiddenError(BaseError, exc.HTTPForbidden):
    pass


class BaseHandler(webapp2.RequestHandler):
    """
    Base class for serializing-deserializing handlers in this module.
    Consider using one of it's descendants.
    """

    def get_request_body(self):
        """
        Returns a converted request body (from JSON, XML, YAML, etc). Should be overridden.
        :raises:
            NotImplementedError
        """
        raise NotImplementedError('BaseHandler is not intended to decode request body')

    def get_returned_mime_type(self):
        """Returns MIME type returned by the handler. Should be overridden.
        :returns:
            MIME type string
        """
        return 'text/plain'

    def prepare_response(self, payload):
        """Prepares object for writing into response (typically, converting it to JSON, XML, YAML etc.
        Should be overridden.

        :param payload:
            payload object
        :returns:
            prepared for ``webapp2.RequestHandler.write()`` object
        """
        return str(payload)

    def respond(self, payload):
        """
        Use this method to talk to the client. Do not override it, consider overriding get_returned_mime_type and
        prepare_response instead.
        """
        self.response.content_type = self.get_returned_mime_type()
        self.response.write(self.prepare_response(payload))

    def handle_exception(self, exception, debug):
        """
        Do not override this method - override `convert_exception` method instead
        """
        resolved_exception = self.resolve_exception(exception, debug)
        self.response.status_int = resolved_exception.code
        self.response.content_type = self.get_returned_mime_type()
        self.respond(self.convert_exception(resolved_exception))

    def convert_exception(self, e):
        """
        Override this method if you want to control how exceptions are represented

        :param e: an exception object
        :return: a serializable dictionary
        """
        return {
            ERROR_CODE: e.code,
            ERROR_TYPE: e.__class__.__name__,
            ERROR_MESSAGE: e.message
        }

    def resolve_exception(self, exception, debug):
        """
        :returns:
            an original exception imbued with a status code or substituted one if not in debug mode
        """
        if isinstance(exception, BaseError):
            e = exception
        else:
            if debug:
                e = exception
                e.code = InternalServerError.code
            else:
                logging.exception(exception)
                e = InternalServerError(exception.message)
        self.response.status_int = e.code
        return e

    def get_bool_param(self, param_name):
        """
        Get boolean query string param

        :param param_name:
            parameter name
        :returns:
            a bool object
        """
        str_param = self.get_str_param(param_name, False)
        return str_param and (str_param.lower() == 'true' or str_param == '1')

    def get_str_param(self, param_name, required=False):
        """
        Get raw query string param

        :param param_name:
            parameter name
        :param required:
            a boolean to enforce requirement. ``False`` by default
        :returns:
            a ``str`` object
        """
        param = self.request.get(param_name)
        if required and not param:
            raise BadRequestError(
                'parameter {} is required'
                .format(param_name)
            )
        return param

    def get_int_param(self, param_name, required=False):
        """
        Get int query string param

        :param param_name:
            parameter name
        :param required:
            a boolean to enforce requirement. ``False`` by default
        :returns:
            an ``int`` object
        """
        str_param = self.get_str_param(param_name, required)
        if not str_param:
            return None
        try:
            param = int(str_param)
            return param
        except Exception:
            raise BadRequestError(
                'parameter {} must be an integer number'
                .format(str_param)
            )

    def get_isodate_param(self, param_name, required=False):
        """
        Get date query string param in ISO 8601 format

        :param param_name:
            parameter name
        :param required:
            a boolean to enforce requirement. ``False`` by default
        :returns:
            a ``datetime.date`` object
        """
        str_param = self.get_str_param(param_name, required)
        if not str_param:
            return None
        try:
            param = datetime.strptime(str_param, '%Y-%m-%d').date()
            return param
        except Exception:
            raise BadRequestError(
                'parameter {} must be a date in ISO format YYYY-mm-dd'
                .format(str_param)
            )

    def get_isodatetime_param(self, param_name, required=False):
        """
        Get datetime query string param in ISO 8601 format

        :param param_name:
            parameter name
        :param required:
            a boolean to enforce requirement. ``False`` by default
        :returns:
            a ``datetime.datetime`` object
        """
        str_param = self.get_str_param(param_name, required)
        if not str_param:
            return None
        try:
            param = datetime.strptime(str_param, '%Y-%m-%dTHH:MM.SS')
            return param
        except Exception:
            raise BadRequestError(
                'parameter {} must be a datetime in ISO format YYYY-MM-DD\'T\'HH:MM.SS'
                .format(str_param)
            )


APPLICATION_JSON = 'application/json'
ERROR_CODE = 'errorCode'
ERROR_TYPE = 'errorType'
ERROR_MESSAGE = 'errorMessage'


class JsonHandler(BaseHandler):
    """
    Handler with JSON serialization
    """

    def get_request_body(self):
        """
        Decodes JSON payload into Python objects

        :returns:
            decoded JSON payload or None
        :raises:
            BadRequestError if payload is not a valid JSON
        """
        if not self.request.body:
            return None
        try:
            payload = json.decode(self.request.body)
        except Exception as e:
            logging.exception(e)
            raise BadRequestError('bad json')
        return payload

    def get_returned_mime_type(self):
        return APPLICATION_JSON

    def prepare_response(self, payload):
        return json.encode(payload)


